from __future__ import annotations

import atexit
import os
import signal
import sys
import threading
import time
from dataclasses import dataclass, field
from types import TracebackType
from typing import Any, Dict, Optional, Type

import httpx

from ._shared import (
    _BaseWyndDataIngestClient,
    _format_ndjson_lines,
    _get_lines_size,
    _is_retryable_error,
    _parse_retry_after_ms,
)
from .types import (
    BatchingOptions,
    CompressionOptions,
    RetryOptions,
    SendOptions,
    WyndDataIngestPayload,
    WyndDataIngestRequestResult,
    WyndDataIngestTableOperation,
)

_SIGNALS = (signal.SIGINT, signal.SIGTERM)
_REGISTERED_CLIENTS: set["WyndDataIngestClient"] = set()
_PROCESS_HOOKS_ATTACHED = False
_ORIGINAL_SIGNAL_HANDLERS: dict[signal.Signals, Any] = {}
_ORIGINAL_EXCEPTHOOK = sys.excepthook
_ORIGINAL_THREADING_EXCEPTHOOK = threading.excepthook


class WyndDataIngestClient(_BaseWyndDataIngestClient):
    def __init__(
        self,
        *,
        url: str,
        connections: int = 16,
        table_name: Optional[str] = None,
        table_operation: Optional[WyndDataIngestTableOperation] = None,
        headers: Optional[Dict[str, str]] = None,
        timeout_ms: int = 30_000,
        user_agent: str = "wynd-dataingest/1.0.0",
        retry: Optional[RetryOptions] = None,
        compression: Optional[CompressionOptions] = None,
        batching: Optional[BatchingOptions] = None,
        client: Optional[httpx.Client] = None,
    ) -> None:
        super().__init__(
            url=url,
            table_name=table_name,
            table_operation=table_operation,
            headers=headers,
            timeout_ms=timeout_ms,
            user_agent=user_agent,
            retry=retry,
            compression=compression,
            batching=batching,
        )
        if client is not None:
            self._client = client
            self._owns_client = False
        else:
            limits = httpx.Limits(
                max_connections=connections,
                max_keepalive_connections=connections,
                keepalive_expiry=10.0,
            )
            self._client = httpx.Client(
                limits=limits,
                timeout=self._default_timeout,
                http2=False,
            )
            self._owns_client = True
        self._batch_condition = threading.Condition()
        self._current_batch = _SyncPendingBatch()
        self._ready_batches: list[_SyncPendingBatch] = []
        self._flush_timer: Optional[threading.Timer] = None
        self._in_flight_flushes = 0

        if self._batching_enabled:
            _register_client_for_process_flush(self)

    def send(
        self,
        payload: WyndDataIngestPayload,
        options: Optional[SendOptions] = None,
    ) -> WyndDataIngestRequestResult:
        if not self._can_batch(payload, options):
            self._drain_pending_batches()
            request_body, headers, timeout = self._prepare_send(payload, options)
            return self._send_serialized(request_body, headers, timeout)

        lines = self._serialize_payload_lines(payload)
        return self._enqueue_batch(lines)

    def close(self) -> None:
        _unregister_client_for_process_flush(self)
        self._clear_flush_timer()
        self._drain_pending_batches()

        if self._owns_client:
            self._client.close()

    def flush_on_process_lifecycle(self) -> None:
        self._clear_flush_timer()
        self._drain_pending_batches()

    def __enter__(self) -> "WyndDataIngestClient":
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self.close()

    def _send_serialized(
        self,
        request_body: bytes,
        headers: Dict[str, str],
        timeout: httpx.Timeout,
    ) -> WyndDataIngestRequestResult:
        attempt = 1
        while True:
            try:
                response = self._client.post(
                    self._url,
                    content=request_body,
                    headers=headers,
                    timeout=timeout,
                )
                if 200 <= response.status_code < 300:
                    return self._build_result(response, attempt)

                retry_after_ms = _parse_retry_after_ms(response.headers.get("retry-after"))
                if (
                    response.status_code not in self._retry.retryable_status_codes
                    or attempt >= self._retry.max_attempts
                ):
                    self._raise_request_error(
                        attempt=attempt,
                        status_code=response.status_code,
                        response_body=response.text,
                    )

                time.sleep(self._compute_delay(attempt, retry_after_ms) / 1000.0)
                attempt += 1
            except httpx.HTTPError as exc:
                if not _is_retryable_error(exc) or attempt >= self._retry.max_attempts:
                    self._raise_request_error(attempt=attempt, cause=exc)

                time.sleep(self._compute_delay(attempt) / 1000.0)
                attempt += 1

    def _can_batch(
        self,
        payload: WyndDataIngestPayload,
        options: Optional[SendOptions],
    ) -> bool:
        if not self._batching_enabled or isinstance(payload, (bytes, bytearray)):
            return False

        if options is None:
            return True

        return (
            options.headers is None
            and options.content_type is None
            and options.timeout_ms is None
        )

    def _enqueue_batch(self, lines: list[str]) -> WyndDataIngestRequestResult:
        deferred = _SyncDeferred()

        with self._batch_condition:
            self._current_batch.lines.extend(lines)
            self._current_batch.bytes += _get_lines_size(lines)
            self._current_batch.deferreds.append(deferred)

            if self._should_flush_current_batch_immediately_locked():
                self._clear_flush_timer_locked()
                self._enqueue_current_batch_for_flush_locked()
                self._dispatch_ready_batches_locked()
            else:
                self._schedule_flush_locked()

        return deferred.wait()

    def _should_flush_current_batch_immediately_locked(self) -> bool:
        return (
            len(self._current_batch.lines) >= self._batching.max_records
            or self._current_batch.bytes >= self._batching.max_bytes
            or self._batching.linger_ms <= 0
        )

    def _schedule_flush_locked(self) -> None:
        if self._flush_timer is not None or not self._current_batch.lines:
            return

        self._flush_timer = threading.Timer(
            self._batching.linger_ms / 1000.0,
            self._flush_timer_callback,
        )
        self._flush_timer.daemon = True
        self._flush_timer.start()

    def _flush_timer_callback(self) -> None:
        with self._batch_condition:
            self._flush_timer = None
            self._enqueue_current_batch_for_flush_locked()
            self._dispatch_ready_batches_locked()

    def _clear_flush_timer(self) -> None:
        with self._batch_condition:
            self._clear_flush_timer_locked()

    def _clear_flush_timer_locked(self) -> None:
        if self._flush_timer is not None:
            self._flush_timer.cancel()
            self._flush_timer = None

    def _enqueue_current_batch_for_flush_locked(self) -> None:
        if not self._current_batch.lines:
            return

        self._ready_batches.append(self._current_batch)
        self._current_batch = _SyncPendingBatch()

    def _dispatch_ready_batches_locked(self) -> None:
        while (
            self._in_flight_flushes < self._batching.max_inflight_batches
            and self._ready_batches
        ):
            batch = self._ready_batches.pop(0)
            self._in_flight_flushes += 1
            worker = threading.Thread(
                target=self._flush_batch_worker,
                args=(batch,),
                daemon=True,
            )
            worker.start()

    def _flush_batch_worker(self, batch: _SyncPendingBatch) -> None:
        try:
            request_body, headers, timeout = self._prepare_serialized_batch(batch.lines)
            result = self._send_serialized(request_body, headers, timeout)
        except BaseException as exc:
            for deferred in batch.deferreds:
                deferred.reject(exc)
        else:
            for deferred in batch.deferreds:
                deferred.resolve(result)
        finally:
            with self._batch_condition:
                self._in_flight_flushes -= 1

                if self._ready_batches:
                    self._dispatch_ready_batches_locked()
                elif self._current_batch.lines:
                    if self._should_flush_current_batch_immediately_locked():
                        self._clear_flush_timer_locked()
                        self._enqueue_current_batch_for_flush_locked()
                        self._dispatch_ready_batches_locked()
                    else:
                        self._schedule_flush_locked()

                self._batch_condition.notify_all()

    def _drain_pending_batches(self) -> None:
        with self._batch_condition:
            self._clear_flush_timer_locked()
            self._enqueue_current_batch_for_flush_locked()
            self._dispatch_ready_batches_locked()

            while self._current_batch.lines or self._ready_batches or self._in_flight_flushes > 0:
                self._batch_condition.wait()

    def _prepare_serialized_batch(
        self,
        lines: list[str],
    ) -> tuple[bytes, Dict[str, str], httpx.Timeout]:
        headers = dict(self._default_headers)
        body = _format_ndjson_lines(lines)
        request_body = self._prepare_request_body(body, headers)
        return request_body, headers, self._default_timeout


def _register_client_for_process_flush(client: WyndDataIngestClient) -> None:
    global _PROCESS_HOOKS_ATTACHED

    _REGISTERED_CLIENTS.add(client)
    if _PROCESS_HOOKS_ATTACHED:
        return

    atexit.register(_handle_atexit)
    sys.excepthook = _handle_uncaught_exception
    threading.excepthook = _handle_thread_exception

    for registered_signal in _SIGNALS:
        _ORIGINAL_SIGNAL_HANDLERS[registered_signal] = signal.getsignal(registered_signal)
        signal.signal(registered_signal, _handle_signal)

    _PROCESS_HOOKS_ATTACHED = True


def _unregister_client_for_process_flush(client: WyndDataIngestClient) -> None:
    _REGISTERED_CLIENTS.discard(client)


def _flush_registered_clients() -> None:
    for client in list(_REGISTERED_CLIENTS):
        try:
            client.flush_on_process_lifecycle()
        except Exception:
            continue


def _handle_atexit() -> None:
    _flush_registered_clients()


def _handle_signal(signum: int, frame: object) -> None:
    _flush_registered_clients()

    previous_handler = _ORIGINAL_SIGNAL_HANDLERS.get(signal.Signals(signum), signal.SIG_DFL)
    if callable(previous_handler) and previous_handler is not _handle_signal:
        previous_handler(signum, frame)
        return

    if previous_handler == signal.SIG_IGN:
        return

    signal.signal(signum, signal.SIG_DFL)
    os.kill(os.getpid(), signum)


def _handle_uncaught_exception(
    exc_type: type[BaseException],
    exc_value: BaseException,
    traceback: TracebackType | None,
) -> None:
    _flush_registered_clients()
    _ORIGINAL_EXCEPTHOOK(exc_type, exc_value, traceback)


def _handle_thread_exception(args: threading.ExceptHookArgs) -> None:
    _flush_registered_clients()
    _ORIGINAL_THREADING_EXCEPTHOOK(args)


@dataclass(slots=True)
class _SyncDeferred:
    _event: threading.Event = field(default_factory=threading.Event)
    _result: Optional[WyndDataIngestRequestResult] = None
    _error: Optional[BaseException] = None

    def resolve(self, value: WyndDataIngestRequestResult) -> None:
        self._result = value
        self._event.set()

    def reject(self, error: BaseException) -> None:
        self._error = error
        self._event.set()

    def wait(self) -> WyndDataIngestRequestResult:
        self._event.wait()
        if self._error is not None:
            raise self._error
        if self._result is None:
            raise RuntimeError("Batch result was not resolved.")
        return self._result


@dataclass(slots=True)
class _SyncPendingBatch:
    lines: list[str] = field(default_factory=list)
    bytes: int = 0
    deferreds: list[_SyncDeferred] = field(default_factory=list)



from .async_client import AsyncWyndDataIngestClient
from .client import WyndDataIngestClient
from .types import (
    BatchingOptions,
    CompressionOptions,
    GzipCompressionOptions,
    RetryOptions,
    SendOptions,
    WyndDataIngestError,
    WyndDataIngestRequestResult,
    WyndDataIngestTableOperation,
)

__all__ = [
    "BatchingOptions",
    "CompressionOptions",
    "GzipCompressionOptions",
    "RetryOptions",
    "SendOptions",
    "AsyncWyndDataIngestClient",
    "WyndDataIngestClient",
    "WyndDataIngestError",
    "WyndDataIngestRequestResult",
    "WyndDataIngestTableOperation",
]



from __future__ import annotations

import asyncio
from dataclasses import dataclass, field
from types import TracebackType
from typing import Dict, Optional, Type

import httpx

from ._shared import (
    _BaseWyndDataIngestClient,
    _format_ndjson_lines,
    _get_lines_size,
    _is_retryable_error,
    _parse_retry_after_ms,
)
from .types import (
    BatchingOptions,
    CompressionOptions,
    RetryOptions,
    SendOptions,
    WyndDataIngestPayload,
    WyndDataIngestRequestResult,
    WyndDataIngestTableOperation,
)


class AsyncWyndDataIngestClient(_BaseWyndDataIngestClient):
    def __init__(
        self,
        *,
        url: str,
        connections: int = 16,
        table_name: Optional[str] = None,
        table_operation: Optional[WyndDataIngestTableOperation] = None,
        headers: Optional[Dict[str, str]] = None,
        timeout_ms: int = 30_000,
        user_agent: str = "wynd-dataingest/1.0.0",
        retry: Optional[RetryOptions] = None,
        compression: Optional[CompressionOptions] = None,
        batching: Optional[BatchingOptions] = None,
        client: Optional[httpx.AsyncClient] = None,
    ) -> None:
        super().__init__(
            url=url,
            table_name=table_name,
            table_operation=table_operation,
            headers=headers,
            timeout_ms=timeout_ms,
            user_agent=user_agent,
            retry=retry,
            compression=compression,
            batching=batching,
        )
        if client is not None:
            self._client = client
            self._owns_client = False
        else:
            limits = httpx.Limits(
                max_connections=connections,
                max_keepalive_connections=connections,
                keepalive_expiry=10.0,
            )
            self._client = httpx.AsyncClient(
                limits=limits,
                timeout=self._default_timeout,
                http2=False,
            )
            self._owns_client = True
        self._current_batch = _AsyncPendingBatch()
        self._ready_batches: list[_AsyncPendingBatch] = []
        self._batch_lock = asyncio.Lock()
        self._flush_handle: Optional[asyncio.TimerHandle] = None
        self._in_flight_flushes: set[asyncio.Task[None]] = set()

    async def send(
        self,
        payload: WyndDataIngestPayload,
        options: Optional[SendOptions] = None,
    ) -> WyndDataIngestRequestResult:
        if not self._can_batch(payload, options):
            await self._drain_pending_batches()
            request_body, headers, timeout = self._prepare_send(payload, options)
            return await self._send_serialized(request_body, headers, timeout)

        lines = self._serialize_payload_lines(payload)
        return await self._enqueue_batch(lines)

    async def aclose(self) -> None:
        self._clear_flush_handle()
        await self._drain_pending_batches()

        if self._owns_client:
            await self._client.aclose()

    async def __aenter__(self) -> "AsyncWyndDataIngestClient":
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        await self.aclose()

    async def _send_serialized(
        self,
        request_body: bytes,
        headers: Dict[str, str],
        timeout: httpx.Timeout,
    ) -> WyndDataIngestRequestResult:
        attempt = 1
        while True:
            try:
                response = await self._client.post(
                    self._url,
                    content=request_body,
                    headers=headers,
                    timeout=timeout,
                )
                if 200 <= response.status_code < 300:
                    return self._build_result(response, attempt)

                retry_after_ms = _parse_retry_after_ms(response.headers.get("retry-after"))
                if (
                    response.status_code not in self._retry.retryable_status_codes
                    or attempt >= self._retry.max_attempts
                ):
                    self._raise_request_error(
                        attempt=attempt,
                        status_code=response.status_code,
                        response_body=response.text,
                    )

                await asyncio.sleep(self._compute_delay(attempt, retry_after_ms) / 1000.0)
                attempt += 1
            except httpx.HTTPError as exc:
                if not _is_retryable_error(exc) or attempt >= self._retry.max_attempts:
                    self._raise_request_error(attempt=attempt, cause=exc)

                await asyncio.sleep(self._compute_delay(attempt) / 1000.0)
                attempt += 1

    def _can_batch(
        self,
        payload: WyndDataIngestPayload,
        options: Optional[SendOptions],
    ) -> bool:
        if not self._batching_enabled or isinstance(payload, (bytes, bytearray)):
            return False

        if options is None:
            return True

        return (
            options.headers is None
            and options.content_type is None
            and options.timeout_ms is None
        )

    async def _enqueue_batch(self, lines: list[str]) -> WyndDataIngestRequestResult:
        loop = asyncio.get_running_loop()
        future: asyncio.Future[WyndDataIngestRequestResult] = loop.create_future()

        async with self._batch_lock:
            self._current_batch.lines.extend(lines)
            self._current_batch.bytes += _get_lines_size(lines)
            self._current_batch.futures.append(future)

            if self._should_flush_current_batch_immediately_locked():
                self._cancel_flush_handle_locked()
                self._enqueue_current_batch_for_flush_locked()
                self._dispatch_ready_batches_locked(loop)
            else:
                self._schedule_flush_locked(loop)

        return await future

    def _should_flush_current_batch_immediately_locked(self) -> bool:
        return (
            len(self._current_batch.lines) >= self._batching.max_records
            or self._current_batch.bytes >= self._batching.max_bytes
            or self._batching.linger_ms <= 0
        )

    def _schedule_flush_locked(self, loop: asyncio.AbstractEventLoop) -> None:
        if self._flush_handle is not None or not self._current_batch.lines:
            return

        self._flush_handle = loop.call_later(
            self._batching.linger_ms / 1000.0,
            self._schedule_flush_task_from_callback,
        )

    def _schedule_flush_task_from_callback(self) -> None:
        self._flush_handle = None
        loop = asyncio.get_running_loop()
        self._enqueue_current_batch_for_flush_locked()
        self._dispatch_ready_batches_locked(loop)

    def _clear_flush_handle(self) -> None:
        if self._flush_handle is not None:
            self._flush_handle.cancel()
            self._flush_handle = None

    def _cancel_flush_handle_locked(self) -> None:
        if self._flush_handle is not None:
            self._flush_handle.cancel()
            self._flush_handle = None

    def _enqueue_current_batch_for_flush_locked(self) -> None:
        if not self._current_batch.lines:
            return

        self._ready_batches.append(self._current_batch)
        self._current_batch = _AsyncPendingBatch()

    def _dispatch_ready_batches_locked(self, loop: asyncio.AbstractEventLoop) -> None:
        while (
            len(self._in_flight_flushes) < self._batching.max_inflight_batches
            and self._ready_batches
        ):
            batch = self._ready_batches.pop(0)
            task = loop.create_task(self._flush_batch_task(batch))
            self._in_flight_flushes.add(task)
            task.add_done_callback(self._handle_flush_task_done)

    async def _flush_batch_task(self, batch: _AsyncPendingBatch) -> None:
        try:
            request_body, headers, timeout = self._prepare_serialized_batch(batch.lines)
            result = await self._send_serialized(request_body, headers, timeout)
        except BaseException as exc:
            for future in batch.futures:
                if not future.done():
                    future.set_exception(exc)
        else:
            for future in batch.futures:
                if not future.done():
                    future.set_result(result)

    def _handle_flush_task_done(self, task: asyncio.Task[None]) -> None:
        self._in_flight_flushes.discard(task)
        asyncio.get_running_loop().create_task(self._after_flush_completion())

    async def _after_flush_completion(self) -> None:
        async with self._batch_lock:
            loop = asyncio.get_running_loop()
            self._dispatch_ready_batches_locked(loop)

            if self._ready_batches:
                return

            if not self._current_batch.lines:
                return

            if self._should_flush_current_batch_immediately_locked():
                self._cancel_flush_handle_locked()
                self._enqueue_current_batch_for_flush_locked()
                self._dispatch_ready_batches_locked(loop)
                return

            self._schedule_flush_locked(loop)

    async def _drain_pending_batches(self) -> None:
        async with self._batch_lock:
            loop = asyncio.get_running_loop()
            self._cancel_flush_handle_locked()
            self._enqueue_current_batch_for_flush_locked()
            self._dispatch_ready_batches_locked(loop)

        while True:
            tasks = list(self._in_flight_flushes)
            if not tasks:
                async with self._batch_lock:
                    if not self._current_batch.lines and not self._ready_batches:
                        return

                    loop = asyncio.get_running_loop()
                    self._enqueue_current_batch_for_flush_locked()
                    self._dispatch_ready_batches_locked(loop)
                    tasks = list(self._in_flight_flushes)
                    if not tasks and not self._current_batch.lines and not self._ready_batches:
                        return

            await asyncio.gather(*tasks, return_exceptions=True)

    def _prepare_serialized_batch(
        self,
        lines: list[str],
    ) -> tuple[bytes, Dict[str, str], httpx.Timeout]:
        headers = dict(self._default_headers)
        body = _format_ndjson_lines(lines)
        request_body = self._prepare_request_body(body, headers)
        return request_body, headers, self._default_timeout


@dataclass(slots=True)
class _AsyncPendingBatch:
    lines: list[str] = field(default_factory=list)
    bytes: int = 0
    futures: list[asyncio.Future[WyndDataIngestRequestResult]] = field(default_factory=list)



from __future__ import annotations

import gzip
import json
import random
import time
from collections.abc import Sequence
from email.utils import parsedate_to_datetime
from typing import Dict, Optional, Tuple, Union

import httpx

from .types import (
    BatchingOptions,
    CompressionOptions,
    GzipCompressionOptions,
    RetryOptions,
    SendOptions,
    WyndDataIngestError,
    WyndDataIngestPayload,
    WyndDataIngestRequestResult,
    WyndDataIngestTableOperation,
)

DEFAULT_GZIP_COMPRESSION = GzipCompressionOptions(min_bytes=100 * 1_024)
DEFAULT_BATCHING = BatchingOptions()


class _BaseWyndDataIngestClient:
    def __init__(
        self,
        *,
        url: str,
        table_name: Optional[str] = None,
        table_operation: Optional[WyndDataIngestTableOperation] = None,
        headers: Optional[Dict[str, str]] = None,
        timeout_ms: int = 30_000,
        user_agent: str = "wynd-dataingest/1.0.0",
        retry: Optional[RetryOptions] = None,
        compression: Optional[CompressionOptions] = None,
        batching: Optional[BatchingOptions] = None,
    ) -> None:
        self._url = url
        self._default_timeout = httpx.Timeout(timeout_ms / 1000.0)
        self._retry = retry or RetryOptions()
        self._compression = compression
        self._batching = _normalize_batching_options(batching)
        self._batching_enabled = batching is not None
        self._default_headers = {
            "content-type": _resolve_content_type(),
            "user-agent": user_agent,
            **_resolve_table_headers(table_name, table_operation),
            **_lowercase_headers(headers),
        }

    def _prepare_send(
        self,
        payload: WyndDataIngestPayload,
        options: Optional[SendOptions] = None,
    ) -> Tuple[bytes, Dict[str, str], httpx.Timeout]:
        body = self._serialize_payload(payload)
        resolved_options = options or SendOptions()
        headers = {
            **self._default_headers,
            **_lowercase_headers(resolved_options.headers),
        }
        if resolved_options.content_type:
            headers["content-type"] = resolved_options.content_type

        request_body = self._prepare_request_body(body, headers)
        timeout = (
            self._default_timeout
            if resolved_options.timeout_ms is None
            else httpx.Timeout(resolved_options.timeout_ms / 1000.0)
        )
        return request_body, headers, timeout

    def _build_result(self, response: httpx.Response, attempt: int) -> WyndDataIngestRequestResult:
        return WyndDataIngestRequestResult(
            status_code=response.status_code,
            ok=True,
            headers=_headers_to_record(response.headers),
            attempts=attempt,
        )

    def _raise_request_error(
        self,
        *,
        attempt: int,
        status_code: Optional[int] = None,
        response_body: Optional[str] = None,
        cause: Optional[BaseException] = None,
    ) -> None:
        if status_code is None:
            raise WyndDataIngestError(
                "WyndDataIngest request failed.",
                attempts=attempt,
                cause=cause,
            ) from cause

        raise WyndDataIngestError(
            f"WyndDataIngest request failed with status {status_code}.",
            attempts=attempt,
            status_code=status_code,
            response_body=response_body,
        )

    def _compute_delay(self, attempt: int, retry_after_ms: Optional[int] = None) -> int:
        if retry_after_ms is not None:
            return min(retry_after_ms, self._retry.max_delay_ms)

        exponential = min(
            int(self._retry.initial_delay_ms * (self._retry.factor ** max(0, attempt - 1))),
            self._retry.max_delay_ms,
        )
        jitter_offset = int(exponential * self._retry.jitter_ratio * random.random())
        return exponential + jitter_offset

    def _serialize_payload(self, payload: WyndDataIngestPayload) -> bytes:
        if isinstance(payload, (bytes, bytearray)):
            return bytes(payload)

        if isinstance(payload, str) or not isinstance(payload, Sequence):
            raise TypeError("NDJSON ingest requires a sequence payload.")

        return _format_ndjson_lines(self._serialize_payload_lines(payload))

    def _serialize_payload_lines(self, payload: WyndDataIngestPayload) -> list[str]:
        if isinstance(payload, (bytes, bytearray)) or isinstance(payload, str) or not isinstance(payload, Sequence):
            raise TypeError("NDJSON ingest batching requires a sequence payload.")

        return [json.dumps(item) for item in payload]

    def _prepare_request_body(self, body: bytes, headers: Dict[str, str]) -> bytes:
        gzip_options = _normalize_gzip_compression(self._compression.gzip if self._compression else None)
        if not gzip_options or "content-encoding" in headers or len(body) < gzip_options.min_bytes:
            return body

        headers["content-encoding"] = "gzip"
        compresslevel = gzip_options.level if gzip_options.level is not None else 9
        return gzip.compress(body, compresslevel=compresslevel)


def _resolve_content_type() -> str:
    return "application/x-ndjson"


def _normalize_gzip_compression(
    value: Union[bool, GzipCompressionOptions, None],
) -> Optional[GzipCompressionOptions]:
    if value is True:
        return DEFAULT_GZIP_COMPRESSION
    if not value:
        return None
    return GzipCompressionOptions(min_bytes=value.min_bytes, level=value.level)


def _normalize_batching_options(value: Optional[BatchingOptions]) -> BatchingOptions:
    if value is None:
        return BatchingOptions(
            linger_ms=DEFAULT_BATCHING.linger_ms,
            max_records=DEFAULT_BATCHING.max_records,
            max_bytes=DEFAULT_BATCHING.max_bytes,
            max_inflight_batches=DEFAULT_BATCHING.max_inflight_batches,
        )

    return BatchingOptions(
        linger_ms=value.linger_ms,
        max_records=value.max_records,
        max_bytes=value.max_bytes,
        max_inflight_batches=max(1, value.max_inflight_batches),
    )


def _resolve_table_headers(
    table_name: Optional[str],
    table_operation: Optional[WyndDataIngestTableOperation],
) -> Dict[str, str]:
    if not table_name:
        return {}

    if table_operation in (None, "insert", "new"):
        return {"table-name": table_name}
    if table_operation == "update":
        return {"table-name": f"{table_name}-updates"}
    return {"table-name": f"{table_name}-deletes"}


def _lowercase_headers(headers: Optional[Dict[str, str]]) -> Dict[str, str]:
    if not headers:
        return {}
    return {key.lower(): value for key, value in headers.items()}


def _headers_to_record(headers: httpx.Headers) -> Dict[str, str]:
    return dict(headers.items())


def _format_ndjson_lines(lines: Sequence[str]) -> bytes:
    return ("\n".join(lines) + "\n").encode("utf-8")


def _get_lines_size(lines: Sequence[str]) -> int:
    return sum(len(line.encode("utf-8")) + 1 for line in lines)


def _parse_retry_after_ms(value: Optional[str]) -> Optional[int]:
    if not value:
        return None

    try:
        return max(0, int(float(value) * 1000))
    except ValueError:
        pass

    try:
        target = parsedate_to_datetime(value).timestamp()
        return max(0, int((target - time.time()) * 1000))
    except Exception:
        return None


def _is_retryable_error(exc: httpx.HTTPError) -> bool:
    return isinstance(
        exc,
        (
            httpx.ConnectError,
            httpx.PoolTimeout,
        ),
    )



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, List, Literal, Optional, Sequence, Union

WyndDataIngestPayloadMode = Literal["ndjson"]
WyndDataIngestTableOperation = Literal["insert", "new", "update", "delete"]
WyndDataIngestPayload = Union[Sequence[object], bytes, bytearray]
HeaderValue = Union[str, List[str]]


@dataclass(slots=True)
class RetryOptions:
    max_attempts: int = 5
    initial_delay_ms: int = 200
    max_delay_ms: int = 5_000
    factor: float = 2.0
    jitter_ratio: float = 0.2
    retryable_status_codes: Sequence[int] = field(
        default_factory=lambda: (408, 425, 429, 500, 502, 503, 504)
    )


@dataclass(slots=True)
class GzipCompressionOptions:
    min_bytes: int
    level: Optional[int] = None


@dataclass(slots=True)
class CompressionOptions:
    gzip: Union[bool, GzipCompressionOptions, None] = None


@dataclass(slots=True)
class BatchingOptions:
    linger_ms: int = 1_000
    max_records: int = 500
    max_bytes: int = 5 * 1_024 * 1_024
    max_inflight_batches: int = 4


@dataclass(slots=True)
class SendOptions:
    headers: Optional[Dict[str, str]] = None
    content_type: Optional[str] = None
    timeout_ms: Optional[int] = None


@dataclass(slots=True)
class WyndDataIngestRequestResult:
    status_code: int
    ok: bool
    headers: Dict[str, HeaderValue]
    attempts: int
    body_text: Optional[str] = None


class WyndDataIngestError(Exception):
    def __init__(
        self,
        message: str,
        *,
        attempts: int,
        status_code: Optional[int] = None,
        response_body: Optional[str] = None,
        cause: Optional[BaseException] = None,
    ) -> None:
        super().__init__(message)
        self.attempts = attempts
        self.status_code = status_code
        self.response_body = response_body
        self.cause = cause



from __future__ import annotations

import asyncio
import concurrent.futures
import gzip
import threading
import time
from http.server import BaseHTTPRequestHandler, ThreadingHTTPServer
from typing import List

from wynd_dataingest import (
    AsyncWyndDataIngestClient,
    BatchingOptions,
    CompressionOptions,
    GzipCompressionOptions,
    RetryOptions,
    SendOptions,
    WyndDataIngestClient,
    WyndDataIngestError,
)


class ScenarioHandler(BaseHTTPRequestHandler):
    statuses: List[int] = [200]
    retry_after: str | None = None
    captured_bodies: List[bytes] = []
    captured_headers: List[dict[str, str]] = []
    response_delay_s: float = 0.0
    active_requests: int = 0
    max_active_requests: int = 0
    lock = threading.Lock()

    def do_POST(self) -> None:  # noqa: N802
        with ScenarioHandler.lock:
            ScenarioHandler.active_requests += 1
            ScenarioHandler.max_active_requests = max(
                ScenarioHandler.max_active_requests,
                ScenarioHandler.active_requests,
            )

        length = int(self.headers.get("content-length", "0"))
        body = self.rfile.read(length)
        ScenarioHandler.captured_bodies.append(body)
        ScenarioHandler.captured_headers.append({k.lower(): v for k, v in self.headers.items()})

        status = ScenarioHandler.statuses.pop(0)
        if ScenarioHandler.response_delay_s > 0:
            time.sleep(ScenarioHandler.response_delay_s)
        self.send_response(status)
        if ScenarioHandler.retry_after is not None:
            self.send_header("Retry-After", ScenarioHandler.retry_after)
        self.end_headers()
        self.wfile.write(b"ok" if status < 400 else b"error")
        with ScenarioHandler.lock:
            ScenarioHandler.active_requests -= 1

    def log_message(self, format: str, *args: object) -> None:  # noqa: A003
        return


def reset_handler(
    statuses: List[int],
    retry_after: str | None = None,
    response_delay_s: float = 0.0,
) -> None:
    ScenarioHandler.statuses = list(statuses)
    ScenarioHandler.retry_after = retry_after
    ScenarioHandler.captured_bodies = []
    ScenarioHandler.captured_headers = []
    ScenarioHandler.response_delay_s = response_delay_s
    ScenarioHandler.active_requests = 0
    ScenarioHandler.max_active_requests = 0


def serve() -> tuple[ThreadingHTTPServer, str]:
    server = ThreadingHTTPServer(("127.0.0.1", 0), ScenarioHandler)
    thread = threading.Thread(target=server.serve_forever, daemon=True)
    thread.start()
    host, port = server.server_address
    return server, f"http://{host}:{port}/"


def test_retries_retryable_responses_with_exponential_backoff() -> None:
    reset_handler([503, 503, 202])
    server, url = serve()
    client = WyndDataIngestClient(
        url=url,
        retry=RetryOptions(max_attempts=3, initial_delay_ms=1, max_delay_ms=2, jitter_ratio=0),
    )

    result = client.send([{"message": "hello"}])

    assert result.ok is True
    assert result.status_code == 202
    assert result.attempts == 3
    assert ScenarioHandler.captured_bodies == [
        b'{"message": "hello"}\n',
        b'{"message": "hello"}\n',
        b'{"message": "hello"}\n',
    ]

    client.close()
    server.shutdown()
    server.server_close()


def test_uses_retry_after_when_present() -> None:
    reset_handler([429, 200], retry_after="0.01")
    server, url = serve()
    client = WyndDataIngestClient(
        url=url,
        retry=RetryOptions(max_attempts=2, initial_delay_ms=1, max_delay_ms=100, jitter_ratio=0),
    )

    start = time.time()
    result = client.send([{"message": "hello"}])
    elapsed_ms = int((time.time() - start) * 1000)

    assert result.ok is True
    assert result.attempts == 2
    assert elapsed_ms >= 10

    client.close()
    server.shutdown()
    server.server_close()


def test_throws_on_non_retryable_errors() -> None:
    reset_handler([400])
    server, url = serve()
    client = WyndDataIngestClient(
        url=url,
        retry=RetryOptions(max_attempts=4, initial_delay_ms=1, max_delay_ms=2, jitter_ratio=0),
    )

    try:
        client.send([{"message": "hello"}])
        assert False, "Expected WyndDataIngestError"
    except WyndDataIngestError as exc:
        assert exc.status_code == 400

    client.close()
    server.shutdown()
    server.server_close()


def test_gzips_payloads_above_threshold() -> None:
    reset_handler([202])
    server, url = serve()
    client = WyndDataIngestClient(
        url=url,
        compression=CompressionOptions(gzip=GzipCompressionOptions(min_bytes=10)),
    )

    payload = [{"message": "this payload should be compressed"}]
    client.send(payload)

    assert ScenarioHandler.captured_headers[0]["content-encoding"] == "gzip"
    assert gzip.decompress(ScenarioHandler.captured_bodies[0]).decode("utf-8") == (
        '{"message": "this payload should be compressed"}\n'
    )

    client.close()
    server.shutdown()
    server.server_close()


def test_skips_gzip_below_threshold() -> None:
    reset_handler([200])
    server, url = serve()
    client = WyndDataIngestClient(
        url=url,
        compression=CompressionOptions(gzip=GzipCompressionOptions(min_bytes=1_000)),
    )

    payload = [{"message": "small"}]
    client.send(payload)

    assert "content-encoding" not in ScenarioHandler.captured_headers[0]
    assert ScenarioHandler.captured_bodies[0].decode("utf-8") == '{"message": "small"}\n'

    client.close()
    server.shutdown()
    server.server_close()


def test_supports_gzip_auto_mode_with_default_options() -> None:
    reset_handler([200])
    server, url = serve()
    client = WyndDataIngestClient(
        url=url,
        compression=CompressionOptions(gzip=True),
    )

    payload = [{"message": "x" * (110 * 1_024)}]
    client.send(payload)

    assert ScenarioHandler.captured_headers[0]["content-encoding"] == "gzip"
    assert gzip.decompress(ScenarioHandler.captured_bodies[0]).decode("utf-8") == (
        '{"message": "' + ("x" * (110 * 1_024)) + '"}\n'
    )

    client.close()
    server.shutdown()
    server.server_close()


def test_requires_sequence_for_default_ndjson_mode() -> None:
    reset_handler([200])
    server, url = serve()
    client = WyndDataIngestClient(url=url)

    try:
        client.send({"message": "hello"})
        assert False, "Expected TypeError"
    except TypeError as exc:
        assert str(exc) == "NDJSON ingest requires a sequence payload."

    client.close()
    server.shutdown()
    server.server_close()


def test_rejects_string_payloads_for_ndjson_ingest() -> None:
    reset_handler([200])
    server, url = serve()
    client = WyndDataIngestClient(url=url)

    try:
        client.send("hello")
        assert False, "Expected TypeError"
    except TypeError as exc:
        assert str(exc) == "NDJSON ingest requires a sequence payload."

    client.close()
    server.shutdown()
    server.server_close()


def test_does_not_retry_on_timeout_to_avoid_duplicate_ingestion() -> None:
    reset_handler([202], response_delay_s=0.5)
    server, url = serve()
    client = WyndDataIngestClient(
        url=url,
        timeout_ms=50,
        retry=RetryOptions(max_attempts=3, initial_delay_ms=1, max_delay_ms=2, jitter_ratio=0),
    )

    try:
        client.send([{"message": "hello"}])
        assert False, "Expected an error"
    except WyndDataIngestError:
        pass

    assert len(ScenarioHandler.captured_bodies) == 1

    client.close()
    server.shutdown()
    server.server_close()


def test_sets_table_name_header_from_table_operation() -> None:
    server, url = serve()
    cases = [
        ("insert", "events"),
        ("new", "events"),
        ("update", "events-updates"),
        ("delete", "events-deletes"),
    ]

    for operation, expected in cases:
        reset_handler([200])
        client = WyndDataIngestClient(url=url, table_name="events", table_operation=operation)  # type: ignore[arg-type]
        client.send([{"message": operation}])
        assert ScenarioHandler.captured_headers[0]["table-name"] == expected
        client.close()

    server.shutdown()
    server.server_close()


def test_supports_pre_serialized_ndjson_bytes() -> None:
    reset_handler([200])
    server, url = serve()
    client = WyndDataIngestClient(url=url)

    payload = b'{"message": "hello"}\n{"message": "world"}\n'
    client.send(payload)

    assert ScenarioHandler.captured_bodies[0] == payload

    client.close()
    server.shutdown()
    server.server_close()


def test_supports_sync_context_manager() -> None:
    reset_handler([200])
    server, url = serve()

    with WyndDataIngestClient(url=url) as client:
        result = client.send([{"message": "hello"}])

    assert result.ok is True
    assert ScenarioHandler.captured_bodies[0] == b'{"message": "hello"}\n'

    server.shutdown()
    server.server_close()


def test_sync_client_does_not_batch_without_batching_options() -> None:
    reset_handler([202, 202])
    server, url = serve()
    client = WyndDataIngestClient(url=url)

    with concurrent.futures.ThreadPoolExecutor(max_workers=2) as executor:
        first = executor.submit(client.send, [{"message": "first"}])
        second = executor.submit(client.send, [{"message": "second"}])
        first_result = first.result()
        second_result = second.result()

    assert first_result.ok is True
    assert second_result.ok is True
    assert len(ScenarioHandler.captured_bodies) == 2
    assert sorted(ScenarioHandler.captured_bodies) == sorted(
        [
            b'{"message": "first"}\n',
            b'{"message": "second"}\n',
        ]
    )

    client.close()
    server.shutdown()
    server.server_close()


def test_sync_client_batches_multiple_sends_within_linger_window() -> None:
    reset_handler([202])
    server, url = serve()
    client = WyndDataIngestClient(
        url=url,
        batching=BatchingOptions(linger_ms=25, max_records=100, max_bytes=10_000),
    )

    with concurrent.futures.ThreadPoolExecutor(max_workers=2) as executor:
        first = executor.submit(client.send, [{"message": "first"}])
        second = executor.submit(client.send, [{"message": "second"}])
        first_result = first.result()
        second_result = second.result()

    assert first_result.ok is True
    assert second_result.ok is True
    assert ScenarioHandler.captured_bodies == [
        b'{"message": "first"}\n{"message": "second"}\n',
    ]

    client.close()
    server.shutdown()
    server.server_close()


def test_sync_client_flushes_pending_batch_on_close() -> None:
    reset_handler([202])
    server, url = serve()
    client = WyndDataIngestClient(
        url=url,
        batching=BatchingOptions(linger_ms=5_000),
    )

    with concurrent.futures.ThreadPoolExecutor(max_workers=1) as executor:
        future = executor.submit(client.send, [{"message": "close-flush"}])
        time.sleep(0.05)
        client.close()
        result = future.result()

    assert result.ok is True
    assert ScenarioHandler.captured_bodies == [
        b'{"message": "close-flush"}\n',
    ]

    server.shutdown()
    server.server_close()


def test_sync_client_flushes_immediately_when_max_records_is_reached() -> None:
    reset_handler([202])
    server, url = serve()
    client = WyndDataIngestClient(
        url=url,
        batching=BatchingOptions(linger_ms=5_000, max_records=2),
    )

    with concurrent.futures.ThreadPoolExecutor(max_workers=2) as executor:
        first = executor.submit(client.send, [{"message": "first"}])
        second = executor.submit(client.send, [{"message": "second"}])
        first_result = first.result()
        second_result = second.result()

    assert first_result.ok is True
    assert second_result.ok is True
    assert ScenarioHandler.captured_bodies == [
        b'{"message": "first"}\n{"message": "second"}\n',
    ]

    client.close()
    server.shutdown()
    server.server_close()


def test_sync_client_can_flush_multiple_batches_concurrently() -> None:
    reset_handler([202, 202], response_delay_s=0.05)
    server, url = serve()
    client = WyndDataIngestClient(
        url=url,
        batching=BatchingOptions(
            linger_ms=0,
            max_records=1,
            max_inflight_batches=2,
        ),
    )

    with concurrent.futures.ThreadPoolExecutor(max_workers=2) as executor:
        first = executor.submit(client.send, [{"message": "first"}])
        second = executor.submit(client.send, [{"message": "second"}])
        first_result = first.result()
        second_result = second.result()

    assert first_result.ok is True
    assert second_result.ok is True
    assert len(ScenarioHandler.captured_bodies) == 2
    assert ScenarioHandler.max_active_requests == 2

    client.close()
    server.shutdown()
    server.server_close()


def test_sync_client_respects_max_inflight_batches_limit() -> None:
    reset_handler([202, 202, 202], response_delay_s=0.05)
    server, url = serve()
    client = WyndDataIngestClient(
        url=url,
        batching=BatchingOptions(
            linger_ms=0,
            max_records=1,
            max_inflight_batches=2,
        ),
    )

    with concurrent.futures.ThreadPoolExecutor(max_workers=3) as executor:
        futures = [
            executor.submit(client.send, [{"message": "first"}]),
            executor.submit(client.send, [{"message": "second"}]),
            executor.submit(client.send, [{"message": "third"}]),
        ]
        results = [future.result() for future in futures]

    assert all(result.ok for result in results)
    assert ScenarioHandler.max_active_requests == 2

    client.close()
    server.shutdown()
    server.server_close()


def test_sync_client_flushes_immediately_when_max_bytes_is_reached() -> None:
    first_event = {"message": "1234567890"}
    second_event = {"message": "abcdefghij"}
    combined_body = b'{"message": "1234567890"}\n{"message": "abcdefghij"}\n'

    reset_handler([202])
    server, url = serve()
    client = WyndDataIngestClient(
        url=url,
        batching=BatchingOptions(linger_ms=5_000, max_bytes=len(combined_body)),
    )

    with concurrent.futures.ThreadPoolExecutor(max_workers=2) as executor:
        first = executor.submit(client.send, [first_event])
        second = executor.submit(client.send, [second_event])
        first_result = first.result()
        second_result = second.result()

    assert first_result.ok is True
    assert second_result.ok is True
    assert ScenarioHandler.captured_bodies == [combined_body]

    client.close()
    server.shutdown()
    server.server_close()


def test_sync_client_drains_pending_batch_before_direct_send_with_request_headers() -> None:
    reset_handler([202, 202])
    server, url = serve()
    client = WyndDataIngestClient(
        url=url,
        batching=BatchingOptions(linger_ms=5_000),
    )

    with concurrent.futures.ThreadPoolExecutor(max_workers=1) as executor:
        first_future = executor.submit(client.send, [{"message": "batched-first"}])
        time.sleep(0.05)
        second_result = client.send(
            [{"message": "direct-second"}],
            options=SendOptions(headers={"x-test-header": "1"}),
        )
        first_result = first_future.result()

    assert first_result.ok is True
    assert second_result.ok is True
    assert ScenarioHandler.captured_bodies == [
        b'{"message": "batched-first"}\n',
        b'{"message": "direct-second"}\n',
    ]
    assert ScenarioHandler.captured_headers[0].get("x-test-header") is None
    assert ScenarioHandler.captured_headers[1]["x-test-header"] == "1"

    client.close()
    server.shutdown()
    server.server_close()


def test_sync_client_propagates_batched_request_failure_to_every_queued_send() -> None:
    reset_handler([400])
    server, url = serve()
    client = WyndDataIngestClient(
        url=url,
        batching=BatchingOptions(linger_ms=5_000, max_records=2),
        retry=RetryOptions(max_attempts=1, initial_delay_ms=1, max_delay_ms=1, jitter_ratio=0),
    )

    with concurrent.futures.ThreadPoolExecutor(max_workers=2) as executor:
        first = executor.submit(client.send, [{"message": "first"}])
        second = executor.submit(client.send, [{"message": "second"}])

        first_error = None
        second_error = None
        try:
            first.result()
        except WyndDataIngestError as exc:
            first_error = exc
        try:
            second.result()
        except WyndDataIngestError as exc:
            second_error = exc

    assert ScenarioHandler.captured_bodies == [
        b'{"message": "first"}\n{"message": "second"}\n',
    ]
    assert first_error is not None
    assert second_error is not None
    assert first_error.status_code == 400
    assert second_error.status_code == 400

    client.close()
    server.shutdown()
    server.server_close()


def test_sync_client_compresses_combined_batched_payload_above_threshold() -> None:
    first_event = {"message": "first-batched-message"}
    second_event = {"message": "second-batched-message"}
    combined_body = b'{"message": "first-batched-message"}\n{"message": "second-batched-message"}\n'

    reset_handler([202])
    server, url = serve()
    client = WyndDataIngestClient(
        url=url,
        batching=BatchingOptions(linger_ms=5_000, max_records=2),
        compression=CompressionOptions(
            gzip=GzipCompressionOptions(min_bytes=len(combined_body))
        ),
    )

    with concurrent.futures.ThreadPoolExecutor(max_workers=2) as executor:
        first = executor.submit(client.send, [first_event])
        second = executor.submit(client.send, [second_event])
        first_result = first.result()
        second_result = second.result()

    assert first_result.ok is True
    assert second_result.ok is True
    assert ScenarioHandler.captured_headers[0]["content-encoding"] == "gzip"
    assert gzip.decompress(ScenarioHandler.captured_bodies[0]) == combined_body

    client.close()
    server.shutdown()
    server.server_close()


def test_sync_client_flushes_pending_batch_on_process_lifecycle_flush() -> None:
    reset_handler([202])
    server, url = serve()
    client = WyndDataIngestClient(
        url=url,
        batching=BatchingOptions(linger_ms=5_000),
    )

    with concurrent.futures.ThreadPoolExecutor(max_workers=1) as executor:
        future = executor.submit(client.send, [{"message": "process-flush"}])
        time.sleep(0.05)
        client.flush_on_process_lifecycle()
        result = future.result()

    assert result.ok is True
    assert ScenarioHandler.captured_bodies == [
        b'{"message": "process-flush"}\n',
    ]

    client.close()
    server.shutdown()
    server.server_close()


def test_async_client_retries_retryable_responses_with_exponential_backoff() -> None:
    async def run_test() -> None:
        reset_handler([503, 503, 202])
        server, url = serve()

        async with AsyncWyndDataIngestClient(
            url=url,
            retry=RetryOptions(max_attempts=3, initial_delay_ms=1, max_delay_ms=2, jitter_ratio=0),
        ) as client:
            result = await client.send([{"message": "hello"}])

        assert result.ok is True
        assert result.status_code == 202
        assert result.attempts == 3
        assert ScenarioHandler.captured_bodies == [
            b'{"message": "hello"}\n',
            b'{"message": "hello"}\n',
            b'{"message": "hello"}\n',
        ]

        server.shutdown()
        server.server_close()

    asyncio.run(run_test())


def test_async_client_supports_async_context_manager() -> None:
    async def run_test() -> None:
        reset_handler([200])
        server, url = serve()

        async with AsyncWyndDataIngestClient(url=url) as client:
            result = await client.send([{"message": "hello"}])

        assert result.ok is True
        assert ScenarioHandler.captured_bodies[0] == b'{"message": "hello"}\n'

        server.shutdown()
        server.server_close()

    asyncio.run(run_test())


def test_async_client_batches_multiple_sends_within_linger_window() -> None:
    async def run_test() -> None:
        reset_handler([202])
        server, url = serve()

        async with AsyncWyndDataIngestClient(
            url=url,
            batching=BatchingOptions(linger_ms=25, max_records=100, max_bytes=10_000),
        ) as client:
            first_result, second_result = await asyncio.gather(
                client.send([{"message": "first"}]),
                client.send([{"message": "second"}]),
            )

        assert first_result.ok is True
        assert second_result.ok is True
        assert ScenarioHandler.captured_bodies == [
            b'{"message": "first"}\n{"message": "second"}\n',
        ]

        server.shutdown()
        server.server_close()

    asyncio.run(run_test())


def test_async_client_can_flush_multiple_batches_concurrently() -> None:
    async def run_test() -> None:
        reset_handler([202, 202], response_delay_s=0.05)
        server, url = serve()

        async with AsyncWyndDataIngestClient(
            url=url,
            batching=BatchingOptions(
                linger_ms=0,
                max_records=1,
                max_inflight_batches=2,
            ),
        ) as client:
            first_result, second_result = await asyncio.gather(
                client.send([{"message": "first"}]),
                client.send([{"message": "second"}]),
            )

        assert first_result.ok is True
        assert second_result.ok is True
        assert len(ScenarioHandler.captured_bodies) == 2
        assert ScenarioHandler.max_active_requests == 2

        server.shutdown()
        server.server_close()

    asyncio.run(run_test())


def test_async_client_respects_max_inflight_batches_limit() -> None:
    async def run_test() -> None:
        reset_handler([202, 202, 202], response_delay_s=0.05)
        server, url = serve()

        async with AsyncWyndDataIngestClient(
            url=url,
            batching=BatchingOptions(
                linger_ms=0,
                max_records=1,
                max_inflight_batches=2,
            ),
        ) as client:
            results = await asyncio.gather(
                client.send([{"message": "first"}]),
                client.send([{"message": "second"}]),
                client.send([{"message": "third"}]),
            )

        assert all(result.ok for result in results)
        assert ScenarioHandler.max_active_requests == 2

        server.shutdown()
        server.server_close()

    asyncio.run(run_test())


def test_async_client_flushes_pending_batch_on_aclose() -> None:
    async def run_test() -> None:
        reset_handler([202])
        server, url = serve()
        client = AsyncWyndDataIngestClient(
            url=url,
            batching=BatchingOptions(linger_ms=5_000),
        )

        send_task = asyncio.create_task(client.send([{"message": "async-close-flush"}]))
        await asyncio.sleep(0.05)
        await client.aclose()
        result = await send_task

        assert result.ok is True
        assert ScenarioHandler.captured_bodies == [
            b'{"message": "async-close-flush"}\n',
        ]

        server.shutdown()
        server.server_close()

    asyncio.run(run_test())


def test_async_client_propagates_batched_request_failure_to_every_queued_send() -> None:
    async def run_test() -> None:
        reset_handler([400])
        server, url = serve()
        client = AsyncWyndDataIngestClient(
            url=url,
            batching=BatchingOptions(linger_ms=5_000, max_records=2),
            retry=RetryOptions(max_attempts=1, initial_delay_ms=1, max_delay_ms=1, jitter_ratio=0),
        )

        results = await asyncio.gather(
            client.send([{"message": "first"}]),
            client.send([{"message": "second"}]),
            return_exceptions=True,
        )

        assert ScenarioHandler.captured_bodies == [
            b'{"message": "first"}\n{"message": "second"}\n',
        ]
        assert len(results) == 2
        assert isinstance(results[0], WyndDataIngestError)
        assert isinstance(results[1], WyndDataIngestError)
        assert results[0].status_code == 400
        assert results[1].status_code == 400

        await client.aclose()
        server.shutdown()
        server.server_close()

    asyncio.run(run_test())


def test_async_client_requires_sequence_payload() -> None:
    async def run_test() -> None:
        reset_handler([200])
        server, url = serve()
        client = AsyncWyndDataIngestClient(url=url)

        try:
            await client.send({"message": "hello"})
            assert False, "Expected TypeError"
        except TypeError as exc:
            assert str(exc) == "NDJSON ingest requires a sequence payload."
        finally:
            await client.aclose()
            server.shutdown()
            server.server_close()

    asyncio.run(run_test())


def test_async_client_throws_on_non_retryable_errors() -> None:
    async def run_test() -> None:
        reset_handler([400])
        server, url = serve()
        client = AsyncWyndDataIngestClient(
            url=url,
            retry=RetryOptions(max_attempts=4, initial_delay_ms=1, max_delay_ms=2, jitter_ratio=0),
        )

        try:
            await client.send([{"message": "hello"}])
            assert False, "Expected WyndDataIngestError"
        except WyndDataIngestError as exc:
            assert exc.status_code == 400
        finally:
            await client.aclose()
            server.shutdown()
            server.server_close()

    asyncio.run(run_test())
